from logging import getLogger
from typing import Optional, Union

from aiogram import Bot
from aiogram.exceptions import TelegramAPIError, TelegramBadRequest
from aiogram.types import (
    CallbackQuery,
    ContentType,
    FSInputFile,
    InputFile,
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
    Message,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
    URLInputFile,
)

from aiogram_dialog.api.entities import (
    MediaAttachment, MediaId, NewMessage, OldMessage, ShowMode,
)
from aiogram_dialog.api.protocols import (
    MessageManagerProtocol, MessageNotModified,
)
from aiogram_dialog.utils import get_media_id

logger = getLogger(__name__)

SEND_METHODS = {
    ContentType.ANIMATION: "send_animation",
    ContentType.AUDIO: "send_audio",
    ContentType.DOCUMENT: "send_document",
    ContentType.PHOTO: "send_photo",
    ContentType.VIDEO: "send_video",
    ContentType.VIDEO_NOTE: "send_video_note",
    ContentType.DICE: "send_dice",
    ContentType.STICKER: "send_sticker",
    ContentType.VOICE: "send_voice",
}

INPUT_MEDIA_TYPES = {
    ContentType.ANIMATION: InputMediaAnimation,
    ContentType.DOCUMENT: InputMediaDocument,
    ContentType.AUDIO: InputMediaAudio,
    ContentType.PHOTO: InputMediaPhoto,
    ContentType.VIDEO: InputMediaVideo,
}

_INVALID_QUERY_ID_MSG = (
    "query is too old and response timeout expired or query id is invalid"
)


def _combine(sent_message: NewMessage, message_result: Message) -> OldMessage:
    media_id = get_media_id(message_result)
    return OldMessage(
        message_id=message_result.message_id,
        chat=message_result.chat,
        has_reply_keyboard=isinstance(
            sent_message.reply_markup, ReplyKeyboardMarkup,
        ),
        text=message_result.text,
        media_uniq_id=(media_id.file_unique_id if media_id else None),
        media_id=(media_id.file_id if media_id else None),
    )


class MessageManager(MessageManagerProtocol):
    async def answer_callback(
            self, bot: Bot, callback_query: CallbackQuery,
    ) -> None:
        try:
            await bot.answer_callback_query(
                callback_query_id=callback_query.id,
            )
        except TelegramAPIError as e:
            if _INVALID_QUERY_ID_MSG in e.message.lower():
                logger.warning("Cannot answer callback: %s", e)
            else:
                raise

    async def get_media_source(
            self, media: MediaAttachment, bot: Bot,
    ) -> Union[InputFile, str]:
        if media.file_id:
            return media.file_id.file_id
        if media.url:
            if media.use_pipe:
                return URLInputFile(media.url, bot=bot)
            return media.url
        else:
            return FSInputFile(media.path)

    def had_media(self, old_message: OldMessage) -> bool:
        return old_message.media_id is not None

    def need_media(self, new_message: NewMessage) -> bool:
        return bool(new_message.media)

    def had_reply_keyboard(self, old_message: Optional[OldMessage]) -> bool:
        if not old_message:
            return False
        return old_message.has_reply_keyboard

    def need_reply_keyboard(self, new_message: Optional[NewMessage]) -> bool:
        if not new_message:
            return False
        return isinstance(new_message.reply_markup, ReplyKeyboardMarkup)

    def _message_changed(
            self, new_message: NewMessage, old_message: OldMessage,
    ) -> bool:
        if new_message.text != old_message.text:
            return True
        # FIXME ????
        if new_message.reply_markup != old_message.has_reply_keyboard:
            return True

        if self.had_media(old_message) != self.need_media(new_message):
            return True
        if not self.need_media(new_message):
            return False
        old_media_id = MediaId(old_message.media_id, old_message.media_uniq_id)
        if new_message.media.file_id != old_media_id:
            return True

        return False

    def _can_edit(self, new_message: NewMessage,
                  old_message: OldMessage) -> bool:
        # we cannot edit message if media appeared or removed
        return (
            self.had_media(old_message) == self.need_media(new_message) and
            not self.had_reply_keyboard(old_message) and
            not self.need_reply_keyboard(new_message)
        )

    async def show_message(
            self, bot: Bot, new_message: NewMessage,
            old_message: Optional[OldMessage],
    ) -> OldMessage:
        if new_message.show_mode is ShowMode.NO_UPDATE:
            raise MessageNotModified("ShowMode is NO_UPDATE")
        if old_message and new_message.show_mode is ShowMode.DELETE_AND_SEND:
            logger.debug(
                "Delete and send new message, because: mode=%s",
                new_message.show_mode,
            )
            # optimize order not to blink
            if self.need_reply_keyboard(new_message):
                sent_message = await self.send_message(bot, new_message)
                await self.remove_message_safe(bot, old_message, new_message)
            else:
                await self.remove_message_safe(bot, old_message, new_message)
                sent_message = await self.send_message(bot, new_message)
            return _combine(new_message, sent_message)
        if not old_message or new_message.show_mode is ShowMode.SEND:
            logger.debug(
                "Send new message, because: mode=%s, has old_message=%s",
                new_message.show_mode,
                bool(old_message),
            )
            await self._remove_kbd(bot, old_message, new_message)
            return _combine(
                new_message,
                await self.send_message(bot, new_message),
            )

        if not self._message_changed(new_message, old_message):
            # nothing changed: text, keyboard or media
            return old_message

        if not self._can_edit(new_message, old_message):
            await self.remove_message_safe(bot, old_message, new_message)
            return _combine(
                new_message,
                await self.send_message(bot, new_message),
            )
        return _combine(
            new_message,
            await self.edit_message_safe(bot, new_message, old_message),
        )

    # Clear
    async def remove_kbd(
            self,
            bot: Bot,
            old_message: Optional[OldMessage],
    ) -> Optional[Message]:
        return await self._remove_kbd(bot, old_message, None)

    async def _remove_kbd(
            self,
            bot: Bot,
            old_message: Optional[OldMessage],
            new_message: Optional[NewMessage],
    ) -> Optional[Message]:
        if self.had_reply_keyboard(old_message):
            if not self.need_reply_keyboard(new_message):
                return await self.remove_reply_kbd(bot, old_message)
        else:
            return await self.remove_inline_kbd(bot, old_message)

    async def remove_inline_kbd(
            self, bot: Bot, old_message: Optional[OldMessage],
    ) -> Optional[Message]:
        if not old_message:
            return
        logger.debug("remove_inline_kbd in %s", old_message.chat)
        try:
            return await bot.edit_message_reply_markup(
                message_id=old_message.message_id,
                chat_id=old_message.chat.id,
            )
        except TelegramBadRequest as err:
            if "message is not modified" in err.message:
                pass  # nothing to remove
            elif "message can't be edited" in err.message:
                pass
            elif "message to edit not found" in err.message:
                pass
            else:
                raise err

    async def remove_reply_kbd(
            self, bot: Bot, old_message: Optional[OldMessage],
    ) -> Optional[Message]:
        if not old_message:
            return
        logger.debug("remove_reply_kbd in %s", old_message.chat)
        try:
            await bot.delete_message(
                chat_id=old_message.chat.id,
                message_id=old_message.message_id,
            )
        except TelegramBadRequest as err:
            if (
                    "message to delete not found" in err.message or
                    "message can't be deleted" in err.message
            ):
                return
        return await self.send_text(
            bot=bot,
            new_message=NewMessage(
                chat=old_message.chat,
                text="...",
                reply_markup=ReplyKeyboardRemove(),
            ),
        )

    async def remove_message_safe(
            self,
            bot: Bot,
            old_message: OldMessage,
            new_message: NewMessage,
    ) -> None:
        try:
            await bot.delete_message(
                chat_id=old_message.chat.id,
                message_id=old_message.message_id,
            )
        except TelegramBadRequest as err:
            if "message to delete not found" in err.message:
                pass
            elif "message can't be deleted" in err.message:
                await self._remove_kbd(bot, old_message, new_message)
            else:
                raise

    # Edit
    async def edit_message_safe(
            self, bot: Bot, new_message: NewMessage, old_message: OldMessage,
    ) -> Message:
        try:
            return await self.edit_message(bot, new_message, old_message)
        except TelegramBadRequest as err:
            if "message is not modified" in err.message:
                raise MessageNotModified from err
            if (
                    "message can't be edited" in err.message or
                    "message to edit not found" in err.message
            ):
                return await self.send_message(bot, new_message)
            else:
                raise

    async def edit_message(
            self, bot: Bot, new_message: NewMessage, old_message: OldMessage,
    ) -> Message:
        if new_message.media:
            if new_message.media.file_id == old_message.media_id:
                return await self.edit_caption(bot, new_message, old_message)
            return await self.edit_media(bot, new_message, old_message)
        else:
            return await self.edit_text(bot, new_message, old_message)

    async def edit_caption(
            self, bot: Bot, new_message: NewMessage, old_message: OldMessage,
    ) -> Message:
        logger.debug("edit_caption to %s", new_message.chat)
        return await bot.edit_message_caption(
            message_id=old_message.message_id,
            chat_id=old_message.chat.id,
            caption=new_message.text,
            reply_markup=new_message.reply_markup,
            parse_mode=new_message.parse_mode,
        )

    async def edit_text(
            self, bot: Bot, new_message: NewMessage, old_message: OldMessage,
    ) -> Message:
        logger.debug("edit_text to %s", new_message.chat)
        return await bot.edit_message_text(
            message_id=old_message.message_id,
            chat_id=old_message.chat.id,
            text=new_message.text,
            reply_markup=new_message.reply_markup,
            parse_mode=new_message.parse_mode,
            disable_web_page_preview=new_message.disable_web_page_preview,
        )

    async def edit_media(
            self, bot: Bot, new_message: NewMessage, old_message: OldMessage,
    ) -> Message:
        logger.debug(
            "edit_media to %s, media_id: %s",
            new_message.chat,
            new_message.media.file_id,
        )
        media = INPUT_MEDIA_TYPES[new_message.media.type](
            caption=new_message.text,
            reply_markup=new_message.reply_markup,
            parse_mode=new_message.parse_mode,
            disable_web_page_preview=new_message.disable_web_page_preview,
            media=await self.get_media_source(new_message.media, bot),
            **new_message.media.kwargs,
        )
        return await bot.edit_message_media(
            message_id=old_message.message_id,
            chat_id=old_message.chat.id,
            media=media,
            reply_markup=new_message.reply_markup,
        )

    # Send
    async def send_message(self, bot: Bot, new_message: NewMessage) -> Message:
        if new_message.media:
            return await self.send_media(bot, new_message)
        else:
            return await self.send_text(bot, new_message)

    async def send_text(self, bot: Bot, new_message: NewMessage) -> Message:
        logger.debug("send_text to %s", new_message.chat)
        return await bot.send_message(
            new_message.chat.id,
            text=new_message.text,
            disable_web_page_preview=new_message.disable_web_page_preview,
            reply_markup=new_message.reply_markup,
            parse_mode=new_message.parse_mode,
        )

    async def send_media(self, bot: Bot, new_message: NewMessage) -> Message:
        logger.debug(
            "send_media to %s, media_id: %s",
            new_message.chat,
            new_message.media.file_id,
        )
        method = getattr(bot, SEND_METHODS[new_message.media.type], None)
        if not method:
            raise ValueError(
                f"ContentType {new_message.media.type} is not supported",
            )
        return await method(
            new_message.chat.id,
            await self.get_media_source(new_message.media, bot),
            caption=new_message.text,
            reply_markup=new_message.reply_markup,
            parse_mode=new_message.parse_mode,
            **new_message.media.kwargs,
        )



import random
import string
import time
from dataclasses import dataclass, field
from typing import List, Optional

from aiogram.fsm.state import State

from aiogram_dialog.api.exceptions import DialogStackOverflow
from .context import Context, Data

DEFAULT_STACK_ID = ""
_STACK_LIMIT = 100
_ID_SYMS = string.digits + string.ascii_letters


def new_int_id() -> int:
    return int(time.time()) % 100000000 + random.randint(0, 99) * 100000000


def id_to_str(int_id: int) -> str:
    if not int_id:
        return _ID_SYMS[0]
    base = len(_ID_SYMS)
    res = ""
    while int_id:
        int_id, mod = divmod(int_id, base)
        res += _ID_SYMS[mod]
    return res


def new_id():
    return id_to_str(new_int_id())


@dataclass(unsafe_hash=True)
class Stack:
    _id: str = field(compare=True, default_factory=new_id)
    intents: List[str] = field(compare=False, default_factory=list)
    last_message_id: Optional[int] = field(compare=False, default=None)
    last_reply_keyboard: bool = field(compare=False, default=False)
    last_media_id: Optional[str] = field(compare=False, default=None)
    last_media_unique_id: Optional[str] = field(compare=False, default=None)
    last_income_media_group_id: Optional[str] = field(
        compare=False, default=None,
    )

    @property
    def id(self):
        return self._id

    def push(self, state: State, data: Data) -> Context:
        if len(self.intents) >= _STACK_LIMIT:
            raise DialogStackOverflow(
                f"Cannot open more dialogs in current stack. "
                f"Max count is {_STACK_LIMIT}",
            )
        context = Context(
            _intent_id=new_id(),
            _stack_id=self.id,
            state=state,
            start_data=data,
        )
        self.intents.append(context.id)
        return context

    def pop(self):
        return self.intents.pop()

    def last_intent_id(self):
        return self.intents[-1]

    def empty(self):
        return not self.intents

    def default(self):
        return self.id == DEFAULT_STACK_ID



from aiogram.types import InlineKeyboardMarkup

from aiogram_dialog import DialogManager
from aiogram_dialog.api.internal.widgets import (
    MarkupFactory, MarkupVariant, RawKeyboard,
)


class InlineKeyboardFactory(MarkupFactory):
    async def render_markup(
            self, data: dict, manager: DialogManager, keyboard: RawKeyboard,
    ) -> MarkupVariant:
        # TODO validate buttons
        return InlineKeyboardMarkup(
            inline_keyboard=keyboard,
        )



from typing import Dict, List

from aiogram_dialog.api.internal.widgets import DataGetter
from aiogram_dialog.api.protocols import DialogManager


class CompositeGetter:
    def __init__(self, *getters: DataGetter):
        self.getters: List[DataGetter] = list(getters)

    async def __call__(self, **kwargs):
        data = {}
        for g in self.getters:
            data.update(await g(**kwargs))
        return data


class StaticGetter:
    def __init__(self, data: Dict):
        self.data = data

    async def __call__(self, **kwargs):
        return self.data


class PreviewAwareGetter:
    def __init__(self, normal_getter: DataGetter, preview_getter: DataGetter):
        self.normal_getter = normal_getter
        self.preview_getter = preview_getter

    async def __call__(self, dialog_manager: DialogManager, **kwargs):
        if dialog_manager.is_preview():
            return await self.preview_getter(
                dialog_manager=dialog_manager, **kwargs,
            )
        else:
            return await self.normal_getter(
                dialog_manager=dialog_manager, **kwargs,
            )



from typing import Any, Awaitable, Callable, Dict, Union

from aiogram import Router
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import TelegramObject, Update

from aiogram_dialog.api.entities import ChatEvent, DialogUpdateEvent
from aiogram_dialog.api.internal import DialogManagerFactory, STORAGE_KEY
from aiogram_dialog.api.protocols import (
    BgManagerFactory, DialogManager, DialogRegistryProtocol,
)

MANAGER_KEY = "dialog_manager"
BG_FACTORY_KEY = "dialog_bg_factory"


class ManagerMiddleware(BaseMiddleware):
    def __init__(
            self,
            dialog_manager_factory: DialogManagerFactory,
            registry: DialogRegistryProtocol,
            router: Router,
    ) -> None:
        super().__init__()
        self.dialog_manager_factory = dialog_manager_factory
        self.registry = registry
        self.router = router

    def _is_event_supported(
            self, event: TelegramObject, data: Dict[str, Any],
    ) -> bool:
        return STORAGE_KEY in data

    async def __call__(
            self,
            handler: Callable[
                [Union[Update, DialogUpdateEvent], Dict[str, Any]],
                Awaitable[Any],
            ],
            event: ChatEvent,
            data: Dict[str, Any],
    ) -> Any:
        if self._is_event_supported(event, data):
            data[MANAGER_KEY] = self.dialog_manager_factory(
                event=event,
                data=data,
                registry=self.registry,
                router=self.router,
            )

        try:
            return await handler(event, data)
        finally:
            manager: DialogManager = data.pop(MANAGER_KEY, None)
            if manager:
                await manager.close_manager()


class BgFactoryMiddleware(BaseMiddleware):
    def __init__(
            self,
            bg_manager_factory: BgManagerFactory,
    ) -> None:
        super().__init__()
        self.bg_manager_factory = bg_manager_factory

    async def __call__(
            self,
            handler: Callable[
                [Union[TelegramObject, DialogUpdateEvent], Dict[str, Any]],
                Awaitable[TelegramObject],
            ],
            event: TelegramObject,
            data: Dict[str, Any],
    ) -> Any:
        data[BG_FACTORY_KEY] = self.bg_manager_factory
        return await handler(event, data)



from abc import abstractmethod
from typing import (
    Any,
    Dict,
    Protocol,
)

from aiogram.fsm.state import State
from aiogram.types import CallbackQuery, Message

from aiogram_dialog.api.entities import NewMessage
from aiogram_dialog.api.protocols import DialogProtocol
from .manager import DialogManager
from .widgets import RawKeyboard


class WindowProtocol(Protocol):
    @abstractmethod
    async def render_text(self, data: Dict,
                          manager: DialogManager) -> str:
        raise NotImplementedError

    @abstractmethod
    async def render_kbd(
            self, data: Dict, manager: DialogManager,
    ) -> RawKeyboard:
        raise NotImplementedError

    @abstractmethod
    async def load_data(
            self,
            dialog: "DialogProtocol",
            manager: DialogManager,
    ) -> Dict:
        raise NotImplementedError

    @abstractmethod
    async def process_message(
            self,
            message: Message,
            dialog: "DialogProtocol",
            manager: DialogManager,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def process_callback(
            self,
            callback: CallbackQuery,
            dialog: "DialogProtocol",
            manager: DialogManager,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def render(
            self,
            dialog: "DialogProtocol",
            manager: DialogManager,
    ) -> NewMessage:
        raise NotImplementedError

    @abstractmethod
    def get_state(self) -> State:
        raise NotImplementedError

    @abstractmethod
    def find(self, widget_id) -> Any:
        raise NotImplementedError



from __future__ import annotations

from abc import abstractmethod
from typing import (
    Any, Callable, Generic, Optional, Protocol, TypeVar, Union,
)

from aiogram.dispatcher.event.handler import FilterObject
from aiogram.types import ContentType, Message

from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import ManagedWidget
from aiogram_dialog.widgets.widget_event import (
    ensure_event_processor,
    WidgetEventProcessor,
)
from .base import BaseInput

T = TypeVar("T")
TypeFactory = Callable[[str], T]


class OnSuccess(Protocol[T]):
    @abstractmethod
    async def __call__(
            self,
            message: Message,
            widget: ManagedTextInput[T],
            dialog_manager: DialogManager,
            data: T,
    ) -> Any:
        raise NotImplementedError


class OnError(Protocol[T]):
    @abstractmethod
    async def __call__(
            self,
            message: Message,
            widget: ManagedTextInput[T],
            dialog_manager: DialogManager,
            error: ValueError,
    ) -> Any:
        raise NotImplementedError


class TextInput(BaseInput, Generic[T]):
    def __init__(
            self,
            id: str,
            type_factory: TypeFactory[T] = str,
            on_success: Union[OnSuccess[T], WidgetEventProcessor, None] = None,
            on_error: Union[OnError, WidgetEventProcessor, None] = None,
            filter: Optional[Callable[..., Any]] = None,
    ):
        super().__init__(id=id)
        if filter is not None:
            self.filter = FilterObject(filter)
        else:
            self.filter = None
        self.type_factory = type_factory
        self.on_success = ensure_event_processor(on_success)
        self.on_error = ensure_event_processor(on_error)

    async def process_message(
            self,
            message: Message,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        if message.content_type != ContentType.TEXT:
            return False
        if self.filter and not await self.filter.call(
                manager.event, **manager.middleware_data,
        ):
            return False
        try:
            value = self.type_factory(message.text)
        except ValueError as err:
            await self.on_error.process_event(
                message, self.managed(manager), manager, err,
            )
        else:
            # store original text
            self.set_widget_data(manager, message.text)
            await self.on_success.process_event(
                message, self.managed(manager), manager, value,
            )
        return True

    def get_value(self, manager: DialogManager) -> T:
        return self.type_factory(self.get_widget_data(manager, None))

    def managed(self, manager: DialogManager):
        return ManagedTextInput(self, manager)


class ManagedTextInput(ManagedWidget[TextInput[T]], Generic[T]):
    def get_value(self) -> T:
        """Get last input data stored by widget."""
        return self.widget.get_value(self.manager)



from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    TypeVar,
    Union,
)

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.internal import RawKeyboard
from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import ManagedWidget, WhenCondition
from aiogram_dialog.widgets.text import Case, Text
from aiogram_dialog.widgets.widget_event import (
    ensure_event_processor,
    WidgetEventProcessor,
)
from .base import Keyboard
from ..common.items import get_items_getter, ItemsGetterVariant

T = TypeVar("T")
ManagedT = TypeVar("ManagedT")
TypeFactory = Callable[[str], T]
ItemIdGetter = Callable[[Any], Union[str, int]]


class OnItemStateChanged(Protocol[ManagedT, T]):
    @abstractmethod
    async def __call__(
            self,
            event: ChatEvent,
            select: ManagedT,  # noqa: F841
            dialog_manager: DialogManager,
            data: T,
    ):
        raise NotImplementedError


class OnItemClick(Protocol[ManagedT, T]):
    @abstractmethod
    async def __call__(
            self,
            event: CallbackQuery,
            select: ManagedT,  # noqa: F841
            dialog_manager: DialogManager,
            data: T,
    ):
        raise NotImplementedError


class Select(Keyboard, Generic[T]):
    def __init__(
            self,
            text: Text,
            id: str,
            item_id_getter: ItemIdGetter,
            items: ItemsGetterVariant,
            type_factory: TypeFactory[T] = str,
            on_click: Union[
                OnItemClick["Select[T]", T], WidgetEventProcessor, None,
            ] = None,
            when: WhenCondition = None,
    ):
        super().__init__(id=id, when=when)
        self.text = text
        self.type_factory = type_factory
        self.on_click = ensure_event_processor(on_click)
        self.item_id_getter = item_id_getter
        self.items_getter = get_items_getter(items)

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> RawKeyboard:
        return [
            [
                await self._render_button(pos, item, data, manager)
                for pos, item in enumerate(self.items_getter(data))
            ],
        ]

    async def _render_button(
            self, pos: int, item: Any, data: Dict,
            manager: DialogManager,
    ) -> InlineKeyboardButton:
        data = {"data": data, "item": item, "pos": pos + 1, "pos0": pos}
        item_id = self.item_id_getter(item)
        return InlineKeyboardButton(
            text=await self.text.render_text(data, manager),
            callback_data=self._item_callback_data(item_id),
        )

    async def _process_item_callback(
            self,
            callback: CallbackQuery,
            data: str,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        await self.on_click.process_event(
            callback,
            self.managed(manager),
            manager,
            self.type_factory(data),
        )
        return True


class StatefulSelect(Select[T], ABC, Generic[T]):
    def __init__(
            self,
            checked_text: Text,
            unchecked_text: Text,
            id: str,
            item_id_getter: ItemIdGetter,
            items: ItemsGetterVariant,
            type_factory: TypeFactory[T] = str,
            on_click: Union[
                OnItemClick[ManagedT, T], WidgetEventProcessor, None,
            ] = None,
            on_state_changed: Union[
                OnItemStateChanged[ManagedT, T], WidgetEventProcessor, None,
            ] = None,
            when: Union[str, Callable] = None,
    ):
        text = Case(
            {True: checked_text, False: unchecked_text},
            selector=self._is_text_checked,
        )
        super().__init__(
            text=text,
            item_id_getter=item_id_getter, items=items,
            on_click=self._process_click,
            id=id, when=when, type_factory=type_factory,
        )
        self.on_item_click = ensure_event_processor(on_click)
        self.on_state_changed = ensure_event_processor(on_state_changed)

    async def _process_on_state_changed(
            self, event: ChatEvent, item_id: str, manager: DialogManager,
    ):
        if self.on_state_changed:
            await self.on_state_changed.process_event(
                event, self.managed(manager), manager,
                self.type_factory(item_id),
            )

    @abstractmethod
    def _is_text_checked(
            self, data: Dict, case: Case, manager: DialogManager,
    ) -> bool:
        raise NotImplementedError

    async def _process_click(
            self,
            callback: CallbackQuery,
            select: ManagedWidget[Select],
            manager: DialogManager,
            item_id: str,
    ):
        if self.on_item_click:
            await self.on_item_click.process_event(
                callback, select, manager, self.type_factory(item_id),
            )
        await self._on_click(callback, select, manager, str(item_id))

    @abstractmethod
    async def _on_click(
            self,
            callback: CallbackQuery,
            select: ManagedWidget[Select],
            manager: DialogManager,
            item_id: str,
    ):
        raise NotImplementedError


class Radio(StatefulSelect[T], Generic[T]):
    def __init__(
            self,
            checked_text: Text,
            unchecked_text: Text,
            id: str,
            item_id_getter: ItemIdGetter,
            items: ItemsGetterVariant,
            type_factory: TypeFactory[T] = str,
            on_click: Union[
                OnItemClick["ManagedRadio[T]", T],
                WidgetEventProcessor, None,
            ] = None,
            on_state_changed: Union[
                OnItemStateChanged["ManagedRadio[T]", T],
                WidgetEventProcessor, None,
            ] = None,
            when: Union[str, Callable] = None,
    ):

        super().__init__(
            checked_text=checked_text,
            unchecked_text=unchecked_text,
            id=id,
            item_id_getter=item_id_getter,
            items=items,
            type_factory=type_factory,
            on_click=on_click,
            on_state_changed=on_state_changed,
            when=when,
        )

    def get_checked(self, manager: DialogManager) -> Optional[str]:
        data = self._get_checked(manager)
        if data is None:
            return None
        return self.type_factory(data)

    def _get_checked(self, manager: DialogManager) -> Optional[str]:
        return self.get_widget_data(manager, None)

    async def set_checked(
            self, event: ChatEvent, item_id: Optional[T],
            manager: DialogManager,
    ):
        checked = self._get_checked(manager)
        item_id_str = str(item_id)
        self.set_widget_data(manager, item_id_str)
        if checked != item_id_str:
            await self._process_on_state_changed(event, item_id_str, manager)

    def is_checked(
            self, item_id: T, manager: DialogManager,
    ) -> bool:
        return str(item_id) == self._get_checked(manager)

    def _preview_checked_id(
            self, manager: DialogManager, item_id: str,
    ) -> str:
        return self.get_widget_data(manager, item_id)

    def _is_text_checked(
            self, data: Dict, case: Case, manager: DialogManager,
    ) -> bool:
        item_id = self.item_id_getter(data["item"])
        if manager.is_preview():
            return item_id == self._preview_checked_id(manager, item_id)
        return self.is_checked(item_id, manager)

    async def _on_click(
            self,
            callback: CallbackQuery,
            select: Select,
            manager: DialogManager,
            item_id: str,
    ):
        await self.set_checked(callback, item_id, manager)

    def managed(self, manager: DialogManager) -> "ManagedRadio[T]":
        return ManagedRadio(self, manager)


class ManagedRadio(ManagedWidget[Radio[T]], Generic[T]):
    def get_checked(self) -> Optional[T]:
        """Get an id of selected item."""
        return self.widget.get_checked(self.manager)

    async def set_checked(self, item_id: T):
        """Get set which item is selected."""
        return await self.widget.set_checked(
            self.manager.event, item_id, self.manager,
        )

    def is_checked(
            self, item_id: T,
    ) -> bool:
        """Get if specified item is checked."""
        return self.widget.is_checked(item_id, self.manager)


class Multiselect(StatefulSelect[T], Generic[T]):
    def __init__(
            self,
            checked_text: Text,
            unchecked_text: Text,
            id: str,
            item_id_getter: ItemIdGetter,
            items: ItemsGetterVariant,
            min_selected: int = 0,
            max_selected: int = 0,
            type_factory: TypeFactory[T] = str,
            on_click: Union[
                OnItemClick["ManagedMultiselect[T]", T],
                WidgetEventProcessor, None,
            ] = None,
            on_state_changed: Union[
                OnItemStateChanged["ManagedMultiselect[T]", T],
                WidgetEventProcessor, None,
            ] = None,
            when: Union[str, Callable] = None,
    ):
        super().__init__(
            checked_text=checked_text,
            unchecked_text=unchecked_text,
            item_id_getter=item_id_getter,
            items=items,
            on_click=on_click,
            on_state_changed=on_state_changed,
            id=id, when=when, type_factory=type_factory,
        )
        self.min_selected = min_selected
        self.max_selected = max_selected

    def _is_text_checked(
            self, data: Dict, case: Case, manager: DialogManager,
    ) -> bool:
        item_id = str(self.item_id_getter(data["item"]))
        if manager.is_preview():
            return (
                # just stupid way to make it differ in preview
                ord(item_id[-1]) % 2 == 1
            )
        return self.is_checked(item_id, manager)

    def is_checked(
            self, item_id: T, manager: DialogManager,
    ) -> bool:
        data = self._get_checked(manager)
        return str(item_id) in data

    def _get_checked(self, manager: DialogManager) -> List[str]:
        return self.get_widget_data(manager, [])

    def get_checked(self, manager: DialogManager) -> List[T]:
        return [self.type_factory(item) for item in self._get_checked(manager)]

    async def reset_checked(
            self, event: ChatEvent, manager: DialogManager,
    ) -> None:
        self.set_widget_data(manager, [])

    async def set_checked(
            self,
            event: ChatEvent,
            item_id: T,
            checked: bool,
            manager: DialogManager,
    ) -> None:
        item_id_str = str(item_id)
        data: List = self._get_checked(manager)
        changed = False
        if item_id_str in data:
            if not checked:
                if len(data) > self.min_selected:
                    data.remove(item_id_str)
                    changed = True
        else:
            if checked:
                if self.max_selected == 0 or self.max_selected > len(data):
                    data.append(item_id_str)
                    changed = True
        if changed:
            self.set_widget_data(manager, data)
            await self._process_on_state_changed(event, item_id_str, manager)

    async def _on_click(
            self,
            callback: CallbackQuery,
            select: Select,
            manager: DialogManager,
            item_id: str,
    ):
        await self.set_checked(
            callback, item_id, not self.is_checked(item_id, manager), manager,
        )

    def managed(self, manager: DialogManager) -> "ManagedMultiselect[T]":
        return ManagedMultiselect(self, manager)


class ManagedMultiselect(ManagedWidget[Multiselect[T]], Generic[T]):
    def is_checked(self, item_id: T) -> bool:
        """Get if an item identified by ``item_id`` is checked."""
        return self.widget.is_checked(item_id, self.manager)

    def get_checked(self) -> List[T]:
        """Get a list of checked items ids."""
        return self.widget.get_checked(self.manager)

    async def reset_checked(self):
        """Reset all items to their default state."""
        return await self.widget.reset_checked(
            self.manager.event, self.manager,
        )

    async def set_checked(self, item_id: T, checked: bool) -> None:
        """Set an item identified by ``item_id`` as checked or unchecked."""
        return await self.widget.set_checked(
            self.manager.event, item_id, checked, self.manager,
        )



__all__ = [
    "DialogProtocol",
    "BaseDialogManager", "BgManagerFactory", "DialogManager",
    "MediaIdStorageProtocol",
    "MessageManagerProtocol", "MessageNotModified",
    "DialogProtocol", "DialogRegistryProtocol",
]

from .dialog import DialogProtocol
from .manager import BaseDialogManager, BgManagerFactory, DialogManager
from .media import MediaIdStorageProtocol
from .message_manager import MessageManagerProtocol, MessageNotModified
from .registry import DialogRegistryProtocol



import html
import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, Type

from aiogram import Router
from aiogram.fsm.state import State, StatesGroup
from aiogram.types import CallbackQuery, Chat, ContentType, Message, User
from jinja2 import Environment, PackageLoader, select_autoescape

from aiogram_dialog import (
    BaseDialogManager, Dialog, DialogManager, DialogProtocol,
)
from aiogram_dialog.api.entities import (
    ChatEvent,
    Context,
    Data,
    DialogAction,
    DialogUpdateEvent,
    MediaAttachment,
    NewMessage,
    ShowMode,
    Stack,
    StartMode,
)
from aiogram_dialog.api.exceptions import NoContextError
from aiogram_dialog.setup import collect_dialogs


@dataclass
class RenderButton:
    state: str
    title: str


@dataclass
class RenderWindow:
    message: str
    state: str
    keyboard: List[List[RenderButton]]
    photo: Optional[str]
    text_input: Optional[RenderButton]
    attachment_input: Optional[RenderButton]


@dataclass
class RenderDialog:
    state_group: str
    windows: List[RenderWindow]


class FakeManager(DialogManager):

    def __init__(self):
        self._event = DialogUpdateEvent(
            from_user=User(id=1, is_bot=False, first_name="Fake"),
            chat=Chat(id=1, type="private"),
            action=DialogAction.UPDATE,
            data={},
            intent_id=None,
            stack_id=None,
        )
        self._context: Optional[Context] = None
        self._dialog: Optional[DialogProtocol] = None
        self._data = {
            "dialog_manager": self,
            "event_chat": Chat(id=1, type="private"),
            "event_from_user": User(id=1, is_bot=False, first_name="Fake"),
        }

    async def next(self) -> None:
        states = self._dialog.states()
        current_index = states.index(self.current_context().state)
        new_state = states[current_index + 1]
        await self.switch_to(new_state)

    async def back(self) -> None:
        states = self._dialog.states()
        current_index = states.index(self.current_context().state)
        new_state = states[current_index - 1]
        await self.switch_to(new_state)

    @property
    def data(self) -> Dict:
        return self._data

    @property
    def event(self) -> ChatEvent:
        return self._event

    async def load_data(self) -> Dict:
        return {}

    async def close_manager(self) -> None:
        pass

    async def reset_stack(self, remove_keyboard: bool = True) -> None:
        self.reset_context()

    def set_dialog(self, dialog: Dialog):
        self._dialog = dialog
        self.reset_context()

    def set_state(self, state: State):
        self.current_context().state = state

    def is_preview(self) -> bool:
        return True

    @property
    def middleware_data(self) -> Dict:
        return {
            "event_chat": self.event.chat,
            "dialog_manager": self,
        }

    @property
    def dialog_data(self) -> Dict:
        return self.current_context().dialog_data

    def reset_context(self) -> None:
        self._context = Context(
            _intent_id="0",
            _stack_id="0",
            start_data={},
            widget_data={},
            dialog_data={},
            state=State(),
        )

    async def switch_to(
            self,
            state: State,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        self.set_state(state)

    async def start(
            self,
            state: State,
            data: Data = None,
            mode: StartMode = StartMode.NORMAL,
            show_mode: ShowMode = ShowMode.AUTO,
    ) -> None:
        self.set_state(state)

    async def done(
            self,
            result: Any = None,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        self.set_state(State("-"))

    def current_stack(self) -> Stack:
        return Stack()

    def current_context(self) -> Context:
        if not self._context:
            raise NoContextError
        return self._context

    def has_context(self) -> bool:
        return bool(self._context)

    async def show_raw(self) -> NewMessage:
        return await self._dialog.render(self)

    async def mark_closed(self) -> None:
        pass

    @property
    def start_data(self) -> Dict:
        return {}

    @property
    def show_mode(self) -> ShowMode:
        return ShowMode.AUTO

    @show_mode.setter
    def show_mode(self, show_mode: ShowMode) -> None:
        return

    async def show(self, show_mode: Optional[ShowMode] = None) -> Message:
        pass

    def find(self, widget_id) -> Optional[Any]:
        widget = self.dialog().find(widget_id)
        if not widget:
            return None
        return widget.managed(self)

    async def update(
            self,
            data: Dict,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        pass

    def bg(
            self,
            user_id: Optional[int] = None, chat_id: Optional[int] = None,
            stack_id: Optional[str] = None, load: bool = False,
    ) -> BaseDialogManager:
        return self

    async def answer_callback(self) -> None:
        pass


def create_photo(media: Optional[MediaAttachment]) -> Optional[str]:
    if not media:
        return None
    if media.type != ContentType.PHOTO:
        return None
    if media.url:
        return media.url
    if media.path:
        return media.path
    if media.file_id:
        return str(media.file_id)
    return None


async def create_button(
        title: str,
        callback: str,
        manager: FakeManager,
        state: State,
        dialog: Dialog,
        simulate_events: bool,
) -> RenderButton:
    if not simulate_events:
        return RenderButton(title=title, state=state.state)
    callback_query = CallbackQuery(
        id="1",
        from_user=User(id=1, is_bot=False, first_name=""),
        chat_instance="",
        data=callback,
    )
    manager.set_state(state)
    try:
        await dialog._callback_handler(callback_query, dialog_manager=manager)
    except Exception:  # noqa: B902
        logging.debug("Click %s", callback)
    state = manager.current_context().state
    return RenderButton(title=title, state=state.state)


async def render_input(
        manager: FakeManager,
        state: State,
        dialog: Dialog,
        content_type: str,
        simulate_events: bool,
) -> Optional[RenderButton]:
    if not simulate_events:
        return None
    if content_type == ContentType.PHOTO:
        data = {content_type: []}
    else:
        data = {content_type: "<stub>"}
    message = Message(
        message_id=1,
        date=datetime.now(),
        chat=Chat(id=1, type="private"),
        **data,
    )
    manager.set_state(state)
    try:
        await dialog._message_handler(message, dialog_manager=manager)
    except Exception:  # noqa: B902
        logging.debug("Input %s", content_type)

    if state == manager.current_context().state:
        logging.debug("State not changed")
        return None
    logging.debug(
        "State changed %s >> %s", state, manager.current_context().state,
    )
    return RenderButton(
        title=content_type,
        state=manager.current_context().state.state,
    )


async def create_window(
        state: State,
        message: NewMessage,
        manager: FakeManager,
        dialog: Dialog,
        simulate_events: bool,
) -> RenderWindow:
    if message.parse_mode is None or message.parse_mode == "None":
        text = html.escape(message.text)
    else:
        text = message.text
    keyboard = []
    for row in message.reply_markup.inline_keyboard:
        keyboard_row = []
        for button in row:
            keyboard_row.append(
                await create_button(
                    title=button.text,
                    callback=button.callback_data,
                    manager=manager,
                    dialog=dialog,
                    state=state,
                    simulate_events=simulate_events,
                ),
            )
        keyboard.append(keyboard_row)

    return RenderWindow(
        message=text.replace("\n", "<br>"),
        state=state.state,
        photo=create_photo(media=message.media),
        keyboard=keyboard,
        text_input=await render_input(
            manager=manager,
            state=state,
            dialog=dialog,
            content_type=ContentType.TEXT,
            simulate_events=simulate_events,
        ),
        attachment_input=await render_input(
            manager=manager,
            state=state,
            dialog=dialog,
            content_type=ContentType.PHOTO,
            simulate_events=simulate_events,
        ),
    )


async def render_dialog(
        manager: FakeManager,
        group: Type[StatesGroup],
        dialog: Dialog,
        simulate_events: bool,
) -> RenderDialog:
    manager.set_dialog(dialog)
    windows = []
    for state in group.__states__:
        manager.set_state(state)
        new_message = await manager.show_raw()
        windows.append(
            await create_window(
                manager=manager,
                state=state,
                dialog=dialog,
                message=new_message,
                simulate_events=simulate_events,
            ),
        )

    return RenderDialog(state_group=str(group), windows=windows)


async def render_preview_content(
        router: Router,
        simulate_events: bool = False,
) -> str:
    fake_manager = FakeManager()
    dialogs = [
        await render_dialog(
            manager=fake_manager,
            group=dialog.states_group(),
            dialog=dialog,
            simulate_events=simulate_events,
        )
        for dialog in collect_dialogs(router)
    ]
    env = Environment(
        loader=PackageLoader("aiogram_dialog.tools"),
        autoescape=select_autoescape(),
    )
    template = env.get_template("message.html")
    return template.render(dialogs=dialogs)


async def render_preview(
        router: Router,
        file: str,
        simulate_events: bool = False,
):
    res = await render_preview_content(router, simulate_events)
    with open(file, "w", encoding="utf-8") as f:
        f.write(res)



from typing import Optional

from aiogram.types import ContentType
from cachetools import LRUCache

from aiogram_dialog.api.entities import MediaId
from aiogram_dialog.api.protocols import MediaIdStorageProtocol


class MediaIdStorage(MediaIdStorageProtocol):
    def __init__(self, maxsize=10240):
        self.cache = LRUCache(maxsize=maxsize)

    async def get_media_id(
            self,
            path: Optional[str],
            url: Optional[str],
            type: ContentType,
    ) -> Optional[MediaId]:
        if not path and not url:
            return None
        return self.cache.get((path, url, type))

    async def save_media_id(
            self,
            path: Optional[str],
            url: Optional[str],
            type: ContentType,
            media_id: MediaId,
    ) -> None:
        if not path and not url:
            return None
        self.cache[(path, url, type)] = media_id



__all__ = [
    "MessageHandlerFunc",
    "MessageInput",
    "BaseInput",
    "CombinedInput",
    "ManagedTextInput",
    "TextInput",
]

from .base import BaseInput, MessageHandlerFunc, MessageInput
from .combined import CombinedInput
from .text import ManagedTextInput, TextInput



from typing import Callable, Dict, Iterable, Optional, Type, Union

from aiogram import Router
from aiogram.dispatcher.event.telegram import TelegramEventObserver
from aiogram.fsm.state import any_state, State, StatesGroup

from aiogram_dialog.api.entities import DIALOG_EVENT_NAME
from aiogram_dialog.api.exceptions import UnregisteredDialogError
from aiogram_dialog.api.internal import DialogManagerFactory
from aiogram_dialog.api.protocols import (
    BgManagerFactory, DialogProtocol, DialogRegistryProtocol,
    MediaIdStorageProtocol, MessageManagerProtocol,
)
from aiogram_dialog.context.intent_middleware import (
    context_saver_middleware,
    IntentErrorMiddleware,
    IntentMiddlewareFactory,
)
from aiogram_dialog.context.media_storage import MediaIdStorage
from aiogram_dialog.manager.bg_manager import BgManagerFactoryImpl
from aiogram_dialog.manager.manager_factory import DefaultManagerFactory
from aiogram_dialog.manager.manager_middleware import (
    BgFactoryMiddleware, ManagerMiddleware,
)
from aiogram_dialog.manager.message_manager import MessageManager
from aiogram_dialog.manager.update_handler import handle_update
from .about import about_dialog


def _setup_event_observer(router: Router) -> None:
    router.observers[DIALOG_EVENT_NAME] = TelegramEventObserver(
        router=router, event_name=DIALOG_EVENT_NAME,
    )


def _register_event_handler(router: Router, callback: Callable) -> None:
    handler = router.observers[DIALOG_EVENT_NAME]
    handler.register(callback, any_state)


class DialogRegistry(DialogRegistryProtocol):
    def __init__(self, router: Router):
        self.router = router
        self._loaded = False
        self._dialogs = {}
        self._state_groups = {}

    def _ensure_loaded(self):
        if not self._loaded:
            self.refresh()

    def find_dialog(self, state: Union[State, str]) -> DialogProtocol:
        self._ensure_loaded()
        try:
            return self._dialogs[state.group]
        except KeyError as e:
            raise UnregisteredDialogError(
                f"No dialog found for `{state.group}`"
                f" (looking by state `{state}`)",
            ) from e

    def state_groups(self) -> Dict[str, Type[StatesGroup]]:
        self._ensure_loaded()
        return self._state_groups

    def refresh(self):
        dialogs = collect_dialogs(self.router)
        self._dialogs = {d.states_group(): d for d in dialogs}
        self._state_groups = {
            d.states_group_name(): d.states_group()
            for d in self._dialogs.values()
        }
        self._loaded = True


def _startup_callback(
        registry: DialogRegistry,
) -> Callable:
    async def _setup_dialogs(router):
        registry.refresh()

    return _setup_dialogs


def _register_middleware(
        router: Router,
        dialog_manager_factory: DialogManagerFactory,
        bg_manager_factory: BgManagerFactory,
):
    registry = DialogRegistry(router)
    manager_middleware = ManagerMiddleware(
        dialog_manager_factory=dialog_manager_factory,
        router=router,
        registry=registry,
    )
    intent_middleware = IntentMiddlewareFactory(registry=registry)
    # delayed configuration of middlewares
    router.startup.register(_startup_callback(registry))
    update_handler = router.observers[DIALOG_EVENT_NAME]

    router.errors.middleware(IntentErrorMiddleware(registry=registry))

    router.message.middleware(manager_middleware)
    router.callback_query.middleware(manager_middleware)
    update_handler.middleware(manager_middleware)
    router.my_chat_member.middleware(manager_middleware)
    router.errors.middleware(manager_middleware)

    router.message.outer_middleware(intent_middleware.process_message)
    router.callback_query.outer_middleware(
        intent_middleware.process_callback_query,
    )
    update_handler.outer_middleware(
        intent_middleware.process_aiogd_update,
    )
    router.my_chat_member.outer_middleware(
        intent_middleware.process_my_chat_member,
    )

    router.message.middleware(context_saver_middleware)
    router.callback_query.middleware(context_saver_middleware)
    update_handler.middleware(context_saver_middleware)
    router.my_chat_member.middleware(context_saver_middleware)

    bg_factory_middleware = BgFactoryMiddleware(bg_manager_factory)
    for observer in router.observers.values():
        observer.outer_middleware(bg_factory_middleware)


def _prepare_dialog_manager_factory(
        dialog_manager_factory: Optional[DialogManagerFactory],
        message_manager: Optional[MessageManagerProtocol],
        media_id_storage: Optional[MediaIdStorageProtocol],
) -> DialogManagerFactory:
    if dialog_manager_factory is not None:
        return dialog_manager_factory
    if media_id_storage is None:
        media_id_storage = MediaIdStorage()
    if message_manager is None:
        message_manager = MessageManager()
    return DefaultManagerFactory(
        message_manager=message_manager,
        media_id_storage=media_id_storage,
    )


def collect_dialogs(router: Router) -> Iterable[DialogProtocol]:
    if isinstance(router, DialogProtocol):
        yield router
    for sub_router in router.sub_routers:
        yield from collect_dialogs(sub_router)


def _include_default_dialogs(router: Router):
    router.include_router(about_dialog())


def setup_dialogs(
        router: Router,
        *,
        dialog_manager_factory: Optional[DialogManagerFactory] = None,
        message_manager: Optional[MessageManagerProtocol] = None,
        media_id_storage: Optional[MediaIdStorageProtocol] = None,
) -> BgManagerFactory:
    _setup_event_observer(router)
    _register_event_handler(router, handle_update)
    _include_default_dialogs(router)

    dialog_manager_factory = _prepare_dialog_manager_factory(
        dialog_manager_factory=dialog_manager_factory,
        message_manager=message_manager,
        media_id_storage=media_id_storage,
    )
    bg_manager_factory = BgManagerFactoryImpl(router)
    _register_middleware(
        router=router,
        dialog_manager_factory=dialog_manager_factory,
        bg_manager_factory=bg_manager_factory,

    )
    return bg_manager_factory



from typing import Optional, Union

from aiogram.types import ContentType

from aiogram_dialog.api.entities import MediaAttachment
from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import WhenCondition
from aiogram_dialog.widgets.text import Const, Text
from .base import Media


class StaticMedia(Media):
    def __init__(
            self,
            *,
            path: Union[Text, str, None] = None,
            url: Union[Text, str, None] = None,
            type: ContentType = ContentType.PHOTO,
            use_pipe: bool = False,
            media_params: dict = None,
            when: WhenCondition = None,
    ):
        super().__init__(when=when)
        if not (url or path):
            raise ValueError("Neither url nor path are provided")
        self.type = type
        if isinstance(path, str):
            path = Const(path)
        self.path = path
        if isinstance(url, str):
            url = Const(url)
        self.url = url
        self.use_pipe = use_pipe
        self.media_params = media_params or {}

    async def _render_media(
            self, data: dict, manager: DialogManager,
    ) -> Optional[MediaAttachment]:
        if self.url:
            url = await self.url.render_text(data, manager)
        else:
            url = None
        if self.path:
            path = await self.path.render_text(data, manager)
        else:
            path = None

        return MediaAttachment(
            type=self.type,
            url=url,
            path=path,
            use_pipe=self.use_pipe,
            **self.media_params,
        )



from typing import Generic, TypeVar

from aiogram_dialog.api.internal import Widget
from aiogram_dialog.api.protocols import DialogManager

W = TypeVar("W", bound=Widget)


class ManagedWidget(Generic[W]):
    def __init__(self, widget: W, manager: DialogManager):
        self.widget = widget
        self.manager = manager

    def __getattr__(self, item):
        return getattr(self.widget, item)
